import { Mapper } from '@automapper/core';
import { InjectMapper } from '@automapper/nestjs';
import { Controller, UsePipes } from '@nestjs/common';
import { CommandBus, QueryBus } from '@nestjs/cqrs';
import { GrpcMethod, RpcException } from '@nestjs/microservices';
import { Territory } from '../../domain/entities/territory';
import { FindAllTerritoriesForPointQuery } from '../../queries/find-all-territories-for-point.query';
import { TerritoryPresenter } from './territory.presenter';
import { ICollection } from '../../../database/src/interfaces/collection.interface';
import { RpcValidationPipe } from '../../../../utils/pipes/rpc.validation-pipe';
import { FindAllTerritoriesForPointRequest } from '../../domain/dtos/find-all-territories-for-point.request';
import { FindAllTerritoriesRequest } from '../../domain/dtos/find-all-territories.request';
import { FindAllTerritoriesQuery } from '../../queries/find-all-territories.query';
import { FindTerritoryByUuidRequest } from '../../domain/dtos/find-territory-by-uuid.request';
import { FindTerritoryByUuidQuery } from '../../queries/find-territory-by-uuid.query';
import { CreateTerritoryRequest } from '../../domain/dtos/create-territory.request';
import { CreateTerritoryCommand } from '../../commands/create-territory.command';
import { DatabaseException } from 'src/modules/database/src/exceptions/database.exception';
import { UpdateTerritoryRequest } from '../../domain/dtos/update-territory.request';
import { UpdateTerritoryCommand } from '../../commands/update-territory.command';
import { DeleteTerritoryCommand } from '../../commands/delete-territory.command';
import { FindAllTerritoriesForPointsRequest } from '../../domain/dtos/find-all-territories-for-points.request';
import { FindAllTerritoriesForPointsQuery } from '../../queries/find-all-territories-for-points.query';

@UsePipes(
  new RpcValidationPipe({
    whitelist: true,
    forbidUnknownValues: false,
  }),
)
@Controller()
export class TerritoriesController {
  constructor(
    private readonly _commandBus: CommandBus,
    private readonly _queryBus: QueryBus,
    @InjectMapper() private readonly _mapper: Mapper,
  ) {}

  @GrpcMethod('TerritoriesService', 'FindAllForPoint')
  async findAllTerritoriesForPoint(
    data: FindAllTerritoriesForPointRequest,
  ): Promise<ICollection<Territory>> {
    const territoryCollection = await this._queryBus.execute(
      new FindAllTerritoriesForPointQuery(data),
    );
    return Promise.resolve({
      data: territoryCollection.data.map((territory: Territory) =>
        this._mapper.map(territory, Territory, TerritoryPresenter),
      ),
      total: territoryCollection.total,
    });
  }

  @GrpcMethod('TerritoriesService', 'FindAllForPoints')
  async findAllTerritoriesForPoints(
    data: FindAllTerritoriesForPointsRequest,
  ): Promise<ICollection<Territory>> {
    const territoryCollection = await this._queryBus.execute(
      new FindAllTerritoriesForPointsQuery(data),
    );
    return Promise.resolve({
      data: territoryCollection.data.map((territory: Territory) =>
        this._mapper.map(territory, Territory, TerritoryPresenter),
      ),
      total: territoryCollection.total,
    });
  }

  @GrpcMethod('TerritoriesService', 'FindAll')
  async findAll(
    data: FindAllTerritoriesRequest,
  ): Promise<ICollection<Territory>> {
    const territoryCollection = await this._queryBus.execute(
      new FindAllTerritoriesQuery(data),
    );
    return Promise.resolve({
      data: territoryCollection.data.map((territory: Territory) =>
        this._mapper.map(territory, Territory, TerritoryPresenter),
      ),
      total: territoryCollection.total,
    });
  }

  @GrpcMethod('TerritoriesService', 'FindOneByUuid')
  async findOneByUuid(
    data: FindTerritoryByUuidRequest,
  ): Promise<TerritoryPresenter> {
    try {
      const territory = await this._queryBus.execute(
        new FindTerritoryByUuidQuery(data),
      );
      return this._mapper.map(territory, Territory, TerritoryPresenter);
    } catch (error) {
      throw new RpcException({
        code: 5,
        message: 'Territory not found',
      });
    }
  }

  @GrpcMethod('TerritoriesService', 'Create')
  async createTerritory(
    data: CreateTerritoryRequest,
  ): Promise<TerritoryPresenter> {
    try {
      const territory = await this._commandBus.execute(
        new CreateTerritoryCommand(data),
      );
      return this._mapper.map(territory, Territory, TerritoryPresenter);
    } catch (e) {
      if (e instanceof DatabaseException) {
        if (e.message.includes('already exists')) {
          throw new RpcException({
            code: 6,
            message: 'Territory already exists',
          });
        }
      }
      throw new RpcException({});
    }
  }

  @GrpcMethod('TerritoriesService', 'Update')
  async updateTerritory(
    data: UpdateTerritoryRequest,
  ): Promise<TerritoryPresenter> {
    try {
      const territory = await this._commandBus.execute(
        new UpdateTerritoryCommand(data),
      );

      return this._mapper.map(territory, Territory, TerritoryPresenter);
    } catch (e) {
      if (e instanceof DatabaseException) {
        if (e.message.includes('not found')) {
          throw new RpcException({
            code: 5,
            message: 'Territory not found',
          });
        }
      }
      throw new RpcException({});
    }
  }

  @GrpcMethod('TerritoriesService', 'Delete')
  async deleteTerritory(data: FindTerritoryByUuidRequest): Promise<void> {
    try {
      await this._commandBus.execute(new DeleteTerritoryCommand(data.uuid));

      return Promise.resolve();
    } catch (e) {
      if (e instanceof DatabaseException) {
        if (e.message.includes('not found')) {
          throw new RpcException({
            code: 5,
            message: 'Territory not found',
          });
        }
      }
      throw new RpcException({});
    }
  }
}
